#include "StartupLocationMapObject.h"

#include "StartupLocationSelection.h"

#include <Map/MapCoordinateSystem.h>
#include <Map/MapEditor.h>
#include <Map/MapProjection.h>
#include <Math/sg_geodesy.h>
#include <Math/CollisionUtils.h>

#include <wx/listimpl.cpp>

#include <map>


WX_DEFINE_LIST(AINetworkVertexNodes);

IMPLEMENT_CLASS(StartupLocationMapObject,AINetworkVertex)

StartupLocationMapObject::StartupLocationMapObject(Airport::SmartRef airport,
                                                   Point3D position,
                                                   float heading,
                                                   wxString gatename,
                                                   wxString gatenumber,
                                                   GateType gatetype,
                                                   wxString airlinecodes,
                                                   float radius):
        AINetworkVertex(airport->GetTransactionManager(),position),
        _airport(airport)
{
        TRANSACTIONOBJECT_STATEVAR(_heading)=heading;
        TRANSACTIONOBJECT_STATEVAR(_gatename)=gatename;
        TRANSACTIONOBJECT_STATEVAR(_gatenumber)=gatenumber;
        TRANSACTIONOBJECT_STATEVAR(_type)=gatetype;
        TRANSACTIONOBJECT_STATEVAR(_airlinecodes)=airlinecodes;
        TRANSACTIONOBJECT_STATEVAR(_radius)=radius;
        _airport->RegisterChangeListener(this);
}

StartupLocationMapObject::~StartupLocationMapObject() {
        _airport->UnregisterChangeListener(this);
}

void StartupLocationMapObject::Paint(wxDC& dc, MapEditor* editor, const MapPass& pass) {
        if (pass!=AINetworkPass) {
                return;
        }
        Point3D localpts[4];
        wxPoint pts[4];
        
        GetLocalCornerPoints(editor,localpts);
        for (int i=0;i<4;i++) {
                pts[i]=editor->GetCoordinateSystem().LogicalToDevice(localpts[i]);
        }
        
        dc.SetPen(*wxBLACK_PEN);
        dc.SetBrush(*wxWHITE_BRUSH);
        
        dc.DrawPolygon(3,pts);
        
        dc.SetBrush(*wxBLACK_BRUSH);
        
        dc.DrawPolygon(3,pts+1);
        
        dc.SetPen(wxNullPen);
        dc.SetBrush(wxNullBrush);
}

bool StartupLocationMapObject::Hit(MapEditor* editor, TDRectangle r) {
        Point3D localpts[4];
        
        GetLocalCornerPoints(editor,localpts);
        return (CollisionUtils::ConvexPolygonRectangleCollide(r,3,localpts) ||
                CollisionUtils::ConvexPolygonRectangleCollide(r,3,localpts+1));
}

MapSelection* StartupLocationMapObject::CreateSelection(MapSelectionManager* manager) {
        return new StartupLocationSelection(manager,this);
}

wxRect StartupLocationMapObject::GetDeviceBounds(MapEditor* editor) {
        Point3D localpts[4];
        wxPoint pt;
        wxCoord left,right,top,bottom;
        
        GetLocalCornerPoints(editor,localpts);
        for (int i=0;i<4;i++) {
                pt=editor->GetCoordinateSystem().LogicalToDevice(localpts[i]);
                if (i==0 || pt.x<left)
                        left=pt.x;
                if (i==0 || pt.x>right)
                        right=pt.x;
                if (i==0 || pt.y<top)
                        top=pt.y;
                if (i==0 || pt.y>bottom)
                        bottom=pt.y;
        }
        return wxRect(left,top,right-left,bottom-top);
}

TDRectangle StartupLocationMapObject::GetGeographicBounds() {
        return TDRectangle(GetPosition());
}

void StartupLocationMapObject::GetLocalCornerPoints(MapEditor* editor, Point3D pts[4]) {
        double r=GetRadius(); // the radius of the parking location in m
        // Determine a lengthwise vector
        double alt,lat1,lon1,az1,lat2,lon2,az2;
        
        lat1=GetPosition().lat();
        lon1=GetPosition().lon();
        az1=GetHeading();
        alt=GetAirport()->GetElevation()*0.3048;
        geo_direct_wgs_84 ( alt, lat1, lon1, az1, 
			      r,&lat2,&lon2,&az2 );
        
        Point3D pos2;
        pos2.setlat(lat2);
        pos2.setlon(lon2);
        
        MapProjection* proj=editor->GetProjection();
        Point3D localpos1,localpos2;
        localpos1=proj->ConvertToLocal(GetPosition());
        localpos2=proj->ConvertToLocal(pos2);
        
        Point3D vx=localpos2-localpos1;
        Point3D vy=Point3D(vx.y(),-vx.x(),0.0);

        // left side, front, back, right side
        pts[0]=localpos1-(8.0*vx+15.0*vy)/17.0;
        pts[1]=localpos2;
        pts[2]=localpos1;
        pts[3]=localpos1-(8.0*vx-15.0*vy)/17.0;
}


void StartupLocationMapObject::clearPushBackRoute()
{
    pushBackNodes.Clear();
}

void StartupLocationMapObject::InsertPushbackPoint(AINetworkVertex* ref)
{
    pushBackNodes.Insert(ref);
}

wxAINetworkVertexNodesNode *StartupLocationMapObject::getFirstPushBackPoint() { 
        /*AINetworkVertexNodes& nodes=TRANSACTIONOBJECT_STATEVAR(nodes);
        wxAINetworkVertexNodesNode* node;
        
        arcNode=node.GetFirst();
        return arcNode;*/
        return pushBackNodes.GetFirst();
};
