#include "CalibrationMode.h"

#include <Map/MapCoordinateSystem.h>
#include <Map/MapEditor.h>
#include <Map/MapProjection.h>
#include <Math/solve.h>

#include <wx/progdlg.h>

#include <cmath>

#ifdef _MSC_VER
# define isnan _isnan
#endif

using namespace std;

IMPLEMENT_CLASS(CalibrationMode,Mode)

BEGIN_EVENT_TABLE(CalibrationMode,Mode)
        EVT_LEFT_DOWN(CalibrationMode::OnLeftDown)
        EVT_LEFT_UP(CalibrationMode::OnLeftUp)
	EVT_LEFT_DCLICK(CalibrationMode::OnLeftDClick)
END_EVENT_TABLE()

CalibrationMode::CalibrationMode(MapCalibratedBackgroundImage::SmartRef bgImg,
                                 bool doRotate,
                                 bool hasFixedScale):
        _bgImg(bgImg),
        _doRotate(doRotate),
        _hasFixedScale(hasFixedScale),
        _acceptClick(false),
        _expectImageClick(true),
        _pointCount(0)
{
}

CalibrationMode::~CalibrationMode() {
}

/* Enter this mode */
bool CalibrationMode::Enter() {
        if (!Mode::Enter())
                return false;
        // Get the fixed scale from the user if required
        if (_hasFixedScale) {
                while (1) {
                        wxTextEntryDialog dialog(NULL, wxT("Enter image scale in meters/pixel"), wxT("Image Scale"), wxT("1"));
                        if (dialog.ShowModal() != wxID_OK)
                                return false;
                        wxString s = dialog.GetValue();
                        if (s.ToDouble(&_fixedScale)) {
                                break;
                        } else if (s.IsEmpty()) {
                                  wxMessageBox(wxT("Please either enter a scale or press cancel to cancel calibration"));
                        } else {
                                  wxMessageBox(wxT("Invalid Entry - please enter a valid number"));
                        }
		}
	}
        wxMessageBox(wxT("Make one click on the background image, followed by one click on the corresponding location of the map.\nRepeat multiple times as required and desired."));
        return true;
}

/* Leave this mode */
void CalibrationMode::Leave(bool force) {
        TransactionManager* tmgr=GetEditor()->GetTransactionManager();
        tmgr->Abort();
        Mode::Leave(force);
}

/* Get a string for the status bar. */
wxString CalibrationMode::GetStatus() {
        if (_expectImageClick) {
                return wxT("Click on a point in the image. Double click for last point.");
        } else {
                return wxT("Click on the corresponding point in the map. Double click for last point.");
        }
}

void CalibrationMode::OnLeftDown(wxMouseEvent& evt) {
        if (evt.AltDown() || evt.ControlDown() || evt.ShiftDown()) {
                evt.Skip();
                return;
        }
        _acceptClick=true;
}

void CalibrationMode::OnLeftUp(wxMouseEvent& evt) {
        if (!_acceptClick || evt.AltDown() || evt.ControlDown() || evt.ShiftDown()) {
                evt.Skip();
                return;
        }
        _acceptClick=false;
        
        MapCoordinateSystem& cs=GetEditor()->GetCoordinateSystem();
        MapProjection* proj=GetEditor()->GetProjection();
        wxPoint devPoint=evt.GetPosition();
        Point3D localPoint=cs.DeviceToLogical(devPoint);
        if (_expectImageClick) {
                double angle=_bgImg->GetAngle();
                // determine the image relative position of this click
                Point3D imageOrigin=proj->ConvertToLocal(_bgImg->GetNorthWestPos());
                
                // offset
                Point3D delta=localPoint-imageOrigin;
                // scale
                delta/=_bgImg->GetScale();
                // rotate and invert y-axis
                Point3D imgDelta;
                
                imgDelta.setx( ::cos(angle)*delta.x()-::sin(angle)*delta.y());
                imgDelta.sety( ::sin(angle)*delta.x()+::cos(angle)*delta.y());
                _imagePoints.push_back(imgDelta);
        } else {
                Point3D geoPoint=proj->ConvertFromLocal(localPoint);
                _mapPoints.push_back(geoPoint);
                _pointCount++;
        }
        _expectImageClick=!_expectImageClick;
        
        UpdateStatus();
}

void CalibrationMode::OnLeftDClick(wxMouseEvent& evt) {
        wxProgressDialog progDlg(wxT("Calibrating image"),
                                 wxT("Calculating calibration parameters"),
                                 0 /* no progress gauge */);
        _sumx=0.0;
        _sumy=0.0;
        _sumu=0.0;
        _sumv=0.0;
        _sumux=0.0;
        _sumuy=0.0;
        _sumvx=0.0;
        _sumvy=0.0;
        _sumxx=0.0;
        _sumyy=0.0;
        _sumxy=0.0;
        _sumuu=0.0;
        _sumvv=0.0;
	// (x,y) is image coordinates, (u,v) is map coordinates
        MapProjection* proj=GetEditor()->GetProjection();
        for (Point3DList::size_type i=0;i<_pointCount;i++) {
                Point3D mapPoint=proj->ConvertToLocal(_mapPoints[i]);
                Point3D imagePoint=_imagePoints[i];
                
                _sumx+=imagePoint.x();
                _sumy+=imagePoint.y();
                _sumu+=mapPoint.x();
                _sumv+=mapPoint.y();
                _sumux+=mapPoint.x()*imagePoint.x();
                _sumuy+=mapPoint.x()*imagePoint.y();
                _sumvx+=mapPoint.y()*imagePoint.x();
                _sumvy+=mapPoint.y()*imagePoint.y();
                _sumxx+=imagePoint.x()*imagePoint.x();
                _sumyy+=imagePoint.y()*imagePoint.y();
                _sumxy+=imagePoint.x()*imagePoint.y();
                _sumuu+=mapPoint.x()*mapPoint.x();
                _sumvv+=mapPoint.y()*mapPoint.y();
        }
        
        if (_doRotate && _hasFixedScale)
                CalculateFixedScaleRotate();
        else if (!_doRotate && _hasFixedScale)
                CalculateFixedScaleNoRotate();
        else if (_doRotate && !_hasFixedScale)
                CalculateUniformScaleRotate();
        else if (!_doRotate && !_hasFixedScale)
                CalculateUniformScaleNoRotate();
        
        // make sure that everything is prepared
        progDlg.Update(-1,wxT("Transforming image"));
        _bgImg->GetImage();
        
        TransactionManager* tmgr=GetEditor()->GetTransactionManager();
        tmgr->Commit(wxT("Calibrating background image"));
        Done();
}

void CalibrationMode::CalculateUniformScaleNoRotate()
{
        if (_pointCount<2) {
                wxMessageBox(wxT("Please specify at least 2 points"));
                Done();
                return;
        }

        /*
         * Uniform scale (sx=sy), no rotation
         *
         * u=S*x+Tx
         * v=S*y+Ty
         *
         * Regression equations:
         *    sum(u*x+v*y-S*(x*x+y*y)-Tx*x-Ty*y) = 0
         *    sum(u      -S*x        -Tx*1-Ty*0) = 0
         *    sum(v      -S*y        -Tx*0-Ty*1) = 0
         */
        double A[3][4]={
                {_sumxx+_sumyy,_sumx      ,_sumy      ,_sumux+_sumvy},
                {_sumx        ,_pointCount,0          ,_sumu        },
                {_sumy        ,0          ,_pointCount,_sumv        }
        };
        double x[3];
        
        solveLinear<double,3>(A,x);
        
        //std::cout << "scale=" << x[0] << " tx=" << x[2] << " ty=" << x[3] << "\n";

        Point3D localNorthWestPos;
        localNorthWestPos.setx(x[1]);
        localNorthWestPos.sety(x[2]);
        
        MapProjection* proj=GetEditor()->GetProjection();
        _bgImg->SetScale(x[0]);
        _bgImg->SetNorthWestPos(proj->ConvertFromLocal(localNorthWestPos));
}

void CalibrationMode::CalculateFixedScaleNoRotate()
{
        if (_pointCount<1) {
                wxMessageBox(wxT("Please specify at least 1 point"));
                Done();
                return;
        }

        /*
         * Fixed scale (scale=sx=sy), no rotation
         *
         * u=scale*x+Tx
         * v=scale*y+Ty
         *
         * Regression equations:
         *    sum(u       -scale*x        -Tx*1-Ty*0) = 0
         *    sum(v       -scale*y        -Tx*0-Ty*1) = 0
         */
        Point3D localNorthWestPos;
        localNorthWestPos.setx((_sumu-_fixedScale*_sumx)/_pointCount);
        localNorthWestPos.sety((_sumv-_fixedScale*_sumy)/_pointCount);
        
        //std::cout << "scale=" << _fixedScale << " tx=" << tx << " ty=" << ty << "\n";

        MapProjection* proj=GetEditor()->GetProjection();
        _bgImg->SetScale(_fixedScale);
        _bgImg->SetNorthWestPos(proj->ConvertFromLocal(localNorthWestPos));
}

void CalibrationMode::CalculateUniformScaleRotate()
{
        if (_pointCount<2) {
                wxMessageBox(wxT("Please specify at least 2 points"));
                Done();
                return;
        }

        /*
         * Uniform scale (sx=sy), with rotation
         *
         * u= A*x+B*y+Tx
         * v=-B*x+A*y+Ty
         *
         * Regression equations:
         *     sum( A*(x*x+y*y)            +Tx*x+Ty*y)=sum(u*x+v*y)
         *     sum(             B*(x*x+y*y)+Tx*y-Ty*x)=sum(u*y-v*x)
         *     sum( A*x        +B*y        +Tx       )=sum(u)
         *     sum( A*y        -B*x             +Ty  )=sum(v)
         */
        double A[4][5]={
                {_sumxx+_sumyy,0            ,_sumx      ,_sumy      ,_sumux+_sumvy},
                {0            ,_sumxx+_sumyy,_sumy      ,-_sumx     ,_sumuy-_sumvx},
                {_sumx        ,_sumy        ,_pointCount,0          ,_sumu        },
                {_sumy        ,-_sumx       ,0          ,_pointCount,_sumv        }
        };
        double x[4];
        
        solveLinear<double,4>(A,x);
        
        double scale=::sqrt(x[0]*x[0]+x[1]*x[1]);
        double angle;
        
                angle=atan2(x[1],x[0]);
        
        //std::cout << "scale=" << scale << " angle=" << angle*180/M_PI << " tx=" << x[2] << " ty=" << x[3] << "\n";
        
        Point3D localNorthWestPos;
        localNorthWestPos.setx(x[2]);
        localNorthWestPos.sety(x[3]);
        
        MapProjection* proj=GetEditor()->GetProjection();
        _bgImg->SetScale(scale);
        _bgImg->SetAngle(angle);
        _bgImg->SetNorthWestPos(proj->ConvertFromLocal(localNorthWestPos));
}

void CalibrationMode::CalculateFixedScaleRotate()
{
        if (_pointCount<2) {
                wxMessageBox(wxT("Please specify at least 2 points"));
                Done();
                return;
        }
	
        /*
         * Fixed scale (scale=sx=sy), with rotation
         *
         * u= A*x+B*y+Tx
         * v=-B*x+A*y+Ty
         * A^2+B^2=scale^2
	 *
	 * Using Lagrange-Multipliers:
	 * grad(e(A,B,C,D))=lambda*grad(A^2+B^2-scale^2)
         *
         * Regression equations:
	 *      sum(A*(x^2+y^2-lambda)                   +C*x+D*y)=sum(u*x+v*y)
	 *      sum(                   B*(x^2+y^2-lambda)+C*y-D*x)=sum(u*y-v*x)
	 *      sum(A*x               +B*y               +C      )=sum(u)
	 *      sum(A*y               -B*x                   +D  )=sum(v)
	 *          A^2               +B^2                        =scale^2
	 *
	 * Solving (eq1-eq3*sum(x)-eq4*sum(y))^2+(eq2-eq3*sum(y)+eq4*sum(x))^2
         *         for lambda after substituting scale^2 for A^2+B^2
         *
	 *        lambda=H+/-sqrt(K)/scale
         *
	 * with H =sum(x^2)-1/n*sum(x)^2+sum(y^2)-1/n*sum(y)^2
	 *      K =Ka^2+Kb^2
	 *      Ka=sum(u*x)-1/n*sum(u)*sum(x)+sum(v*y)-1/n*sum(v)*sum(y)
	 *      Kb=sum(u*y)-1/n*sum(u)*sum(y)-sum(v*x)+1/n*sum(v)*sum(x)
	 *
	 * With this we can solve the linear equation and select the solution
         * with the smaller error.
	 *
         */
	double H,K,Ka,Kb,lambda1,lambda2;
	H =_sumxx-_sumx*_sumx/_pointCount+_sumyy-_sumy*_sumy/_pointCount;
	Ka=_sumux-_sumu*_sumx/_pointCount+_sumvy-_sumv*_sumy/_pointCount;
	Kb=_sumuy-_sumu*_sumy/_pointCount-_sumvx+_sumv*_sumx/_pointCount;
	K=Ka*Ka+Kb*Kb;
	lambda1=H+::sqrt(K)/_fixedScale;
	lambda2=H-::sqrt(K)/_fixedScale;
        
        //std::cout << "H=" << H << " sqrt(K)/scale=" << ::sqrt(K)/_fixedScale << "\n";
        //std::cout << "lambda1=" << lambda1 << " lambda2=" << lambda2 << "\n";
        
	double A1[4][5]={
		{_sumxx+_sumyy-lambda1,0                    ,_sumx      ,_sumy      ,_sumux+_sumvy},
		{0                    ,_sumxx+_sumyy-lambda1,_sumy      ,-_sumx     ,_sumuy-_sumvx},
		{_sumx                ,_sumy                ,_pointCount,0          ,_sumu},
		{_sumy                ,-_sumx               ,0          ,_pointCount,_sumv}
	};
	double A2[4][5]={
		{_sumxx+_sumyy-lambda2,0                    ,_sumx      ,_sumy      ,_sumux+_sumvy},
		{0                    ,_sumxx+_sumyy-lambda2,_sumy      ,-_sumx     ,_sumuy-_sumvx},
		{_sumx                ,_sumy                ,_pointCount,0          ,_sumu},
		{_sumy                ,-_sumx               ,0          ,_pointCount,_sumv}
	};

        double x1[4],x2[4],*x;
        double angle;
        
        solveLinear<double,4>(A1,x1);
        solveLinear<double,4>(A2,x2);

	/* find the solution with the smaller error */
	double e1,e2;
	e1=CalculateSigma( x1[0],x1[1],x1[2],
			  -x1[1],x1[0],x1[3]);
	e2=CalculateSigma( x2[0],x2[1],x2[2],
			  -x2[1],x2[0],x2[3]);
        
        //std::cout << "e1=" << e1 << " e2=" << e2 << "\n";
        
        if (e1<e2 || isnan(e2))
		x=x1;
	else
		x=x2;
        
        if (fabs(x[0])>fabs(x[1])) {
                angle=atan2(x[1],x[0]);
        } else {
                angle=M_PI/2-atan2(x[0],x[1]);
        }

        //std::cout << "scale=" << ::sqrt(x[0]*x[0]+x[1]*x[1]) << " angle=" << angle*180/M_PI << " tx=" << x[2] << " ty=" << x[3] << "\n";

        Point3D localNorthWestPos;
        localNorthWestPos.setx(x[2]);
        localNorthWestPos.sety(x[3]);
        
        MapProjection* proj=GetEditor()->GetProjection();
        _bgImg->SetScale(_fixedScale);
        _bgImg->SetAngle(angle);
        _bgImg->SetNorthWestPos(proj->ConvertFromLocal(localNorthWestPos));
}

double CalibrationMode::CalculateSigma(double A, double B, double C,
				       double D, double E, double F) {
	double ex,ey;
	ex=A*A*_sumxx/_pointCount+B*B*_sumyy/_pointCount+C*C+_sumuu/_pointCount
		+2*A*B*_sumxy/_pointCount+2*A*C*_sumx/_pointCount-2*A*_sumux/_pointCount
		                         +2*B*C*_sumy/_pointCount-2*B*_sumuy/_pointCount
		                                                 -2*C*_sumu/_pointCount;
	ey=D*D*_sumxx/_pointCount+E*E*_sumyy/_pointCount+F*F+_sumvv/_pointCount
		+2*D*E*_sumxy/_pointCount+2*D*F*_sumx/_pointCount-2*D*_sumvx/_pointCount
		                         +2*E*F*_sumy/_pointCount-2*E*_sumvy/_pointCount
		                                                 -2*F*_sumv/_pointCount;
	return (ex+ey);
}

