#include "MapSimpleBackgroundImageSelection.h"
#include "MapSimpleBackgroundImage.h"
#include "MapSimpleBackgroundImageCornerHandle.h"
#include <Map/MapEditor.h>
#include <Map/MapProjection.h>
#include <Map/MapSelectionManager.h>

IMPLEMENT_CLASS(MapSimpleBackgroundImageSelection,MapSelection);

MapSimpleBackgroundImageSelection::MapSimpleBackgroundImageSelection(MapSelectionManager* manager, MapSimpleBackgroundImage* content):
        MapSelection(manager,content) {
}

MapSimpleBackgroundImageSelection::~MapSimpleBackgroundImageSelection() {
}

MapSelectionHandle* MapSimpleBackgroundImageSelection::HitHandle(Point3D p) {
        if (!HandlesActive())
                return (MapSelectionHandle*)NULL;
        MapEditor* editor=GetSelectionManager()->GetEditor();
        Point3D corners[4];
        GetContent()->GetLocalCorners(editor,corners);
        Point3D handlesize=GetHandleBoxLogicalSize();
        for (unsigned int i=0;i<4;i++) {
                TDRectangle handlebounds=TDRectangle(corners[i],handlesize.x(),handlesize.y());
                if (handlebounds.Contains(p)) {
                        return new MapSimpleBackgroundImageCornerHandle(this,i);
                }
        }
        return (MapSelectionHandle*)NULL;
}

wxRect MapSimpleBackgroundImageSelection::GetDeviceBounds() const {
        wxRect r=GetContent()->GetDeviceBounds(GetEditor());
        if (HandlesActive()) {
                wxSize handleSize=GetHandleBoxDeviceSize();
                
                wxPoint corners[4];
                GetContent()->GetDeviceCorners(GetEditor(),corners);
                
                for (unsigned int i=0;i<4;i++) {
                        wxRect handleRect=wxRect(corners[i].x-handleSize.x/2,
                                                 corners[i].y-handleSize.y/2,
                                                 handleSize.x, handleSize.y);
                        r=r.Union(handleRect);
                }
        }
        return r;
}

void MapSimpleBackgroundImageSelection::DeltaDrag(Point3D delta) {
        MapEditor* editor=GetSelectionManager()->GetEditor();
        MapProjection* proj=editor->GetProjection();
        Point3D localNorthWestPos=proj->ConvertToLocal(GetContent()->GetNorthWestPos());
        localNorthWestPos+=delta;
        GetContent()->SetNorthWestPos(proj->ConvertFromLocal(localNorthWestPos));
}

wxString MapSimpleBackgroundImageSelection::EndDrag() {
        return wxT("Moving background image");
}

void MapSimpleBackgroundImageSelection::Paint(wxDC& dc) {
        MapEditor* editor=GetSelectionManager()->GetEditor();
        wxPoint corners[4];
        GetContent()->GetDeviceCorners(editor,corners);

        wxPen pen=wxPen(Selection_Marking_Colour,1,wxDOT);
        dc.SetPen(pen);
        dc.SetBrush(*wxTRANSPARENT_BRUSH);
        dc.DrawPolygon(4,corners);
        dc.SetBrush(wxNullBrush);
        dc.SetPen(wxNullPen);
        if (HandlesActive()) {
                PaintHandles(dc);
        }
}

void MapSimpleBackgroundImageSelection::ProvidePropertiesForClass(wxPropertyGrid* propgrid, wxClassInfo* targetClass) {
        MapSelection::ProvidePropertiesForClass(propgrid,targetClass);
        if (!targetClass->IsKindOf(CLASSINFO(MapSimpleBackgroundImageSelection)))
                return;
        
        propgrid->Append(wxFileProperty(wxT("File"), wxT("MapSimpleBackgroundImageSelection_File")));
        propgrid->Append(wxFloatProperty(wxT("North Latitude"), wxT("MapSimpleBackgroundImageSelection_North")));
        propgrid->Append(wxFloatProperty(wxT("West Longitude"), wxT("MapSimpleBackgroundImageSelection_West")));
        propgrid->Append(wxFloatProperty(wxT("Scale"), wxT("MapSimpleBackgroundImageSelection_Scale")));
}

void MapSimpleBackgroundImageSelection::UpdatePropertiesForClass(wxPropertyGrid* propgrid, wxClassInfo* targetClass) {
        MapSelection::UpdatePropertiesForClass(propgrid,targetClass);
        if (!targetClass->IsKindOf(CLASSINFO(MapSimpleBackgroundImageSelection)))
                return;

        Point3D northWestPos=GetContent()->GetNorthWestPos();
        propgrid->SetPropertyValue(wxT("MapSimpleBackgroundImageSelection_File"),GetContent()->GetFilename());
        propgrid->SetPropertyValue(wxT("MapSimpleBackgroundImageSelection_North"),northWestPos.lat());
        propgrid->SetPropertyValue(wxT("MapSimpleBackgroundImageSelection_West"),northWestPos.lon());
        propgrid->SetPropertyValue(wxT("MapSimpleBackgroundImageSelection_Scale"),GetContent()->GetScale());
}

void MapSimpleBackgroundImageSelection::MarkDifferingPropertiesForClass(wxPropertyGrid* propgrid, wxClassInfo* targetClass, MapSelection* firstSelection) {
        MapSelection::MarkDifferingPropertiesForClass(propgrid,targetClass,firstSelection);
        if (!targetClass->IsKindOf(CLASSINFO(MapSimpleBackgroundImageSelection)))
                return;
        MapSimpleBackgroundImage* firstContent=wxDynamicCast(firstSelection,MapSimpleBackgroundImageSelection)->GetContent();
        MapSimpleBackgroundImage* myContent=GetContent();

        Point3D firstNorthWestPos=firstContent->GetNorthWestPos();
        Point3D myNorthWestPos=myContent->GetNorthWestPos();
        if (firstContent->GetFilename().Cmp(myContent->GetFilename())!=0)
                propgrid->SetPropertyUnspecified(wxT("MapSimpleBackgroundImageSelection_File"));
        if (firstNorthWestPos.lat()!=myNorthWestPos.lat())
                propgrid->SetPropertyUnspecified(wxT("MapSimpleBackgroundImageSelection_North"));
        if (firstNorthWestPos.lon()!=myNorthWestPos.lon())
                propgrid->SetPropertyUnspecified(wxT("MapSimpleBackgroundImageSelection_West"));
        if (firstContent->GetScale()!=myContent->GetScale())
                propgrid->SetPropertyUnspecified(wxT("MapSimpleBackgroundImageSelection_Scale"));
}

void MapSimpleBackgroundImageSelection::UpdateObjectFromProperty(wxPGProperty* property) {
        const wxString& name=property->GetName();
        wxVariant value=property->GetValueAsVariant();

        if (name.Cmp(wxT("MapSimpleBackgroundImageSelection_File"))==0) {
                GetContent()->SetFilename(property->GetValueAsString());
        } else if (name.Cmp(wxT("MapSimpleBackgroundImageSelection_North"))==0) {
                Point3D nwPos=GetContent()->GetNorthWestPos();
                nwPos.setlat(value.GetDouble());
                GetContent()->SetNorthWestPos(nwPos);
        } else if (name.Cmp(wxT("MapSimpleBackgroundImageSelection_West"))==0) {
                Point3D nwPos=GetContent()->GetNorthWestPos();
                nwPos.setlon(value.GetDouble());
                GetContent()->SetNorthWestPos(nwPos);
        } else if (name.Cmp(wxT("MapSimpleBackgroundImageSelection_Scale"))==0) {
                GetContent()->SetScale(value.GetDouble());
        } else {
                MapSelection::UpdateObjectFromProperty(property);
        }
}

void MapSimpleBackgroundImageSelection::PaintHandles(wxDC& dc) {
        MapEditor* editor=GetSelectionManager()->GetEditor();
        wxSize handleSize=GetHandleBoxDeviceSize();
        
        wxPoint corners[4];
        GetContent()->GetDeviceCorners(editor,corners);
        
        wxPen handlePen=wxPen(Selection_Handle_Colour);
        wxBrush handleBrush=wxBrush(Selection_Handle_Colour);
        dc.SetPen(handlePen);
        dc.SetBrush(handleBrush);
        for (unsigned int i=0;i<4;i++) {
                dc.DrawRectangle(corners[i].x-handleSize.x/2,
                                 corners[i].y-handleSize.y/2,
                                 handleSize.x, handleSize.y);
        }
        dc.SetBrush(wxNullBrush);
        dc.SetPen(wxNullPen);
}


