#if HAVE_CONFIG_H
#  include <config.h>
#endif /* HAVE_CONFIG_H */

#include "MapObjectContainer.h"
#include "MapObject.h"
#include "MapEditor.h"

IMPLEMENT_CLASS(MapObjectContainer,MapLayer);

MapObjectContainer::MapObjectContainer(TransactionManager* tmgr, wxString name):
        MapLayer(tmgr,name) {
}

MapObjectContainer::~MapObjectContainer() {
        MapObjectList::Node *node;
        for (node=_objects.GetFirst();node;node=node->GetNext())
                node->GetData()->TransactionObject::UnRef();
}

/* Paint this layer using the given device context into the given editor. */
void MapObjectContainer::Paint(wxDC& dc, MapEditor* editor, const MapPass& pass) {
        MapObjectList::Node* objNode;
        wxCoord x,y,w,h;
        dc.GetClippingBox(&x,&y,&w,&h);
        wxRect cliprect=wxRect(x,y,w,h);
        wxRect devrect;
        for (objNode=_objects.GetFirst();objNode;objNode=objNode->GetNext()) {
                MapObject* obj=objNode->GetData();
                devrect=obj->GetDeviceBounds(editor);
                if (!cliprect.Intersects(devrect))
                        continue;
                obj->Paint(dc,editor,pass);
        }
}

/* Add all objects in this layer intersecting the given hit-TDRectangle r.
 * r is given in the local coordinate system of editor. */
void MapObjectContainer::GetHitObjects(MapObjectList& list,
                                       MapEditor* editor,
                                       TDRectangle r,
                                       bool includeIntersected) {
        MapObjectList::Node* objNode;
        
        if (includeIntersected) {
                for (objNode=_objects.GetFirst();objNode;objNode=objNode->GetNext())
                        if (objNode->GetData()->Hit(editor,r))
                                list.Append(objNode->GetData());
        } else {
                for (objNode=_objects.GetFirst();objNode;objNode=objNode->GetNext())
                        if (r.Contains(objNode->GetData()->GetLocalBounds(editor)))
                                list.Append(objNode->GetData());
        }
}

/* Return the device bounds in the given editor */
wxRect MapObjectContainer::GetDeviceBounds(MapEditor* editor) {
        MapObjectList::Node* objNode;
        wxRect r;
        
        for (objNode=_objects.GetFirst();objNode;objNode=objNode->GetNext()) {
                wxRect devRect=objNode->GetData()->GetDeviceBounds(editor);
                r=r.Union(devRect);
        }
        return r;
}

/* Return the local bounds in the given editor */
TDRectangle MapObjectContainer::GetLocalBounds(MapEditor* editor) {
        MapObjectList::Node* objNode;
        TDRectangle r;
        for (objNode=_objects.GetFirst();objNode;objNode=objNode->GetNext())
                r=TDRectangle::Union(r,objNode->GetData()->GetLocalBounds(editor));
        return r;
}

/* Return the geographic (lat/lon) bounds */
TDRectangle MapObjectContainer::GetGeographicBounds() {
        MapObjectList::Node* objNode;
        TDRectangle r;
        for (objNode=_objects.GetFirst();objNode;objNode=objNode->GetNext())
                r=TDRectangle::Union(r,objNode->GetData()->GetGeographicBounds());
        return r;
}

void MapObjectContainer::GetObjectsOfClass(MapObjectList& list, wxClassInfo* targetClass) {
        MapObjectList::Node* objNode;
        for (objNode=_objects.GetFirst();objNode;objNode=objNode->GetNext()) {
                if (objNode->GetData()->IsKindOf(targetClass))
                        list.Append(objNode->GetData());
        }
}

void MapObjectContainer::AttachObject(MapObject* obj) {
        wxASSERT_MSG(obj->GetContainer()==NULL,
                        wxT("Object must be detached before it can be reattached"));
        BeforeWrite();
        obj->TransactionObject::Ref();
        _objects.Append(obj);
        obj->SetContainer(this);
        obj->Damage();
        AfterWrite();
}

void MapObjectContainer::AttachObject(MapObject* obj, size_t index) {
        wxASSERT_MSG(obj->GetContainer()==NULL,
                        wxT("Object must be detached before it can be reattached"));
        BeforeWrite();
        obj->TransactionObject::Ref();
        _objects.Insert(index, obj);
        obj->SetContainer(this);
        obj->Damage();
        AfterWrite();
}

void MapObjectContainer::DetachObject(MapObject* obj) {
        wxASSERT_MSG(obj->GetContainer()==this,
                     wxT("Object cannot be detached from container it is not attached to"));
        if (_objects.Find(obj)==NULL)
                return;
        BeforeWrite();
        obj->Damage();
        _objects.DeleteObject(obj);
        obj->OnDetach();
        obj->SetContainer(NULL);
        obj->TransactionObject::UnRef();
        AfterWrite();
}

void MapObjectContainer::RestoreMemento(wxObject* object) {
        MapLayer::RestoreMemento(object);
        Memento* memento=dynamic_cast<Memento*>(object);
        wxASSERT(memento);
        
        MapLayer::Damage(); /* Force redraw in old boundaries */
        MapObjectList::Node *node;
        for (node=memento->_objects.GetFirst();node;node=node->GetNext()) {
                node->GetData()->_container=this;
                node->GetData()->Ref();
        }
        for (node=_objects.GetFirst();node;node=node->GetNext()) {
                node->GetData()->UnRef();
        }
        _objects=memento->_objects;
        MapLayer::Damage(); /* Force redraw in new boundaries */
}

void MapObjectContainer::MoveObjectUp(MapObject* obj) {
        wxASSERT_MSG(obj->GetContainer()==this,
                     wxT("Cannot move an object which is not contained in this container"));
        MapObjectList::Node* node=_objects.Find(obj);
        wxASSERT_MSG(node!=(MapObjectList::Node*)NULL,
                     wxT("Object not found in this container, although it should be attached"));
        if (node==_objects.GetLast())
                return;
        MapObjectList::Node* nextNode=node->GetNext();
        
        BeforeWrite();
        if (nextNode==_objects.GetLast())
                _objects.Append(obj);
        else
                _objects.Insert(nextNode->GetNext(),obj);
#       ifdef __WXDEBUG__
                // agh, otherwise the compiler complains about unused var
                bool ret=
#       endif
        _objects.DeleteNode(node);
        wxASSERT(ret);
        
        Damage(obj);
        AfterWrite();
}

void MapObjectContainer::MoveObjectDown(MapObject* obj) {
        wxASSERT_MSG(obj->GetContainer()==this,
                     wxT("Cannot move an object which is not contained in this container"));
        MapObjectList::Node* node=_objects.Find(obj);
        wxASSERT_MSG(node!=(MapObjectList::Node*)NULL,
                     wxT("Object not found in this container, although it should be attached"));
        if (node==_objects.GetFirst())
                return;
        
        BeforeWrite();
        _objects.Insert(node->GetPrevious(),obj);
#       ifdef __WXDEBUG__
                // agh, otherwise the compiler complains about unused var
                bool ret=
#       endif
        _objects.DeleteNode(node);
        wxASSERT(ret);
        
        Damage(obj);
        AfterWrite();
}

void MapObjectContainer::MoveObjectToTop(MapObject* obj) {
        wxASSERT_MSG(obj->GetContainer()==this,
                     wxT("Cannot move an object which is not contained in this container"));
        MapObjectList::Node* node=_objects.Find(obj);
        wxASSERT_MSG(node!=(MapObjectList::Node*)NULL,
                     wxT("Object not found in this container, although it should be attached"));
        
        BeforeWrite();
#       ifdef __WXDEBUG__
                // agh, otherwise the compiler complains about unused var
                bool ret=
#       endif
        _objects.DeleteNode(node);
        wxASSERT(ret);
        _objects.Append(obj);
        
        Damage(obj);
        AfterWrite();
}

void MapObjectContainer::MoveObjectToBottom(MapObject* obj) {
        wxASSERT_MSG(obj->GetContainer()==this,
                     wxT("Cannot move an object which is not contained in this container"));
        BeforeWrite();
#       ifdef __WXDEBUG__
                // agh, otherwise the compiler complains about unused var
                bool ret=
#       endif
        _objects.DeleteObject(obj);
        wxASSERT_MSG(ret,
                     wxT("Object not found in this container, although it should be attached"));
        _objects.Insert(obj);
        
        Damage(obj);
        AfterWrite();
}

void MapObjectContainer::FlipObjectsOrder() {
        MapObjectList newList;
        
        MapObjectList::Node* node;
        
        for (node=_objects.GetFirst();node;node=node->GetNext()) {
                newList.Insert(node->GetData());
        }
        
        BeforeWrite();
        _objects=newList;
        MapLayer::Damage();
        AfterWrite();
}

void MapObjectContainer::Damage(MapObject* obj) {
        MapLayerEditors::Node* editorNode;
        for (editorNode=GetEditors().GetFirst();
             editorNode;
             editorNode=editorNode->GetNext()) {
               wxRect r=obj->GetDeviceBounds(editorNode->GetData());
               editorNode->GetData()->DamageRect(r);
        }
}

