#ifndef MapObjectPropertyManager_h
#define MapObjectPropertyManager_h

// For compilers that support precompilation, includes "wx.h".
#include <wx/wxprec.h>

#ifdef __BORLANDC__
#pragma hdrstop
#endif

#ifndef WX_PRECOMP
// Include your minimal set of headers here, or wx.h
#include <wx/wx.h>
#endif

#include "MapPropertyManager.h"

/**
 * \brief A MapObjectPropertyManager manages a property grid, handling property
 * changes and updates for MapObject instances.
 *
 * There can be multiple property managers for each MapEditor, each managing
 * their own property grid.
 *
 * \note The property manager must be destroyed manually before destroying the
 * associated MapEditor object.
 */

class MapObjectPropertyManager: public MapPropertyManager,
                                public MapSelectionChangeListener {
        
        DECLARE_CLASS(MapObjectPropertyManager)
        
public:
	MapObjectPropertyManager(MapEditor* editor, wxPropertyGrid* grid);
        virtual ~MapObjectPropertyManager();
        
        /**
         * \brief Update the property grid.
         *
         * This method updates the managed property grid using three of the four
         * property management methods in MapSelection:
         *
         * - MapSelection::ProvidePropertiesForClass()
         * - MapSelection::UpdatePropertiesForClass()
         * - MapSelection::MarkDifferingPropertiesForClass()
         *
         * When the selection has changed since the last update, the grid is
         * cleared and the most specific common superclass of all selections
         * is determined using MapSelectionManager::FindCommonSelectionSuperClass().
         *
         * A representative for this class is then found using
         * MapSelectionManager::FindFirstSelectionOfType() and
         * MapSelection::ProvidePropertiesForClass() is called to populate the
         * grid.
         *
         * On each update event MapSelection::UpdatePropertiesForClass() is called on
         * the first selection to provide values for the properties.
         *
         * This is followed by iterating all further selection objects, calling
         * MapSelection::MarkDifferingPropertiesForClass() on them to mark
         * properties with differing values for the selected objects appropriately.
         */
        virtual void UpdateProperties();
        
        /**
         * \brief React on a property change.
         *
         * This is the event handler for the EVT_PG_CHANGED event from the
         * property grid.
         *
         * It iterates over all currently present selections and calls
         * MapSelection::UpdateObjectFromProperty() on each of them,
         * passing the modified property as parameter.
         */
        virtual void PropertyChanged(wxPropertyGridEvent& evt);

        /**
         * \brief React on the addition of selections to the selection manager.
         *
         * This simply schedules a property update event using SchedulePropertyUpdate()
         * and registers this property manager with the selected objects as
         * object change listener.
         */
        virtual void SelectionsAdded(MapSelectionManager* mgr, const MapSelectionList& sellist);
        
        /**
         * \brief React on the removal of selections from the selection manager.
         *
         * This simply schedules a property update event using SchedulePropertyUpdate()
         * and unregisters this property manager from the selected objects as
         * object change listener.
         */
        virtual void SelectionsRemoved(MapSelectionManager* mgr, const MapSelectionList& sellist);
protected:
        MapEditor* _editor;
        bool _selectionChanged;
};

#endif

