#ifndef MapSelectionManager_h
#define MapSelectionManager_h

// For compilers that support precompilation, includes "wx.h".
#include <wx/wxprec.h>

#ifdef __BORLANDC__
#pragma hdrstop
#endif

#ifndef WX_PRECOMP
// Include your minimal set of headers here, or wx.h
#include <wx/wx.h>
#endif

#include <Transactions/TransactionObject.h>

#include "MapDraggable.h"
#include "MapSelection.h"
#include "MapSelectionChangeListener.h"

class MapEditor;
class MapObject;

WX_DECLARE_LIST(MapSelectionChangeListener,MapSelectionChangeListenerList);
WX_DECLARE_VOIDPTR_HASH_MAP(MapSelection*,MapSelectionMap);

/**
 * \brief A MapSelectionManager manages the list of MapSelection instances in
 * a MapEditor.
 *
 * It provides services to manipulate the list of selections, paint selection
 * markers. It also provides basic support to determine the provider of
 * popup menu actions for a given set of selections.
 */
class MapSelectionManager: public TransactionObject,
                        public wxEvtHandler,
                        public MapDraggable {
        
        DECLARE_CLASS(MapSelectionManager)
        TRANSACTIONOBJECT_DEFINE_SMARTREF(MapSelectionManager);
        
public:
	MapSelectionManager(MapEditor* editor);
        virtual ~MapSelectionManager();
        
        /**
         * \brief Add a selection for an object to the list.
         *
         * \return the new selection instance
         */
        MapSelection* AddSelectionFor(MapObject* obj);
        
        /**
         * \brief Remove a given selection from the list.
         */
        void RemoveSelection(MapSelection* selection);
        
        /**
         * \brief Remove all selections from the list.
         */
        void DeselectAll();
        
        /**
         * \brief Get the current selection for a given object.
         *
         * \return The selection for the object or NULL if the object is not selected.
         */
        MapSelection* GetSelectionForObject(MapObject* obj);
        
        /**
         * \brief Paint all selections in the list.
         *
         * \param dc    The device context to paint on.
         */
        void Paint(wxDC& dc);
        
        /**
         * \brief Schedule all selections for repaint.
         */
        void Damage();
        
        /**
         * \brief Determine whether a given logical point lies within any of
         * the present selections.
         *
         * \param p     The point in logical coordinates to check for intersection
         *              with any of the present selections.
         *
         * \return true if and only if the points hits any selection.
         */
        bool Hit(Point3D p);
        
        /**
         * \brief Determine whether a given logical points hits a handle of
         *        one of the present selections.
         *
         * If more than one handle is hit, it is undefined, which of the hit
         * handles is returned.
         *
         * \param p     The point in logical coordinates to check for intersection
         *              with a handle of one of the present selections.
         *
         * \return the selection handle object, if one was hit, NULL otherwise.
         */
        MapSelectionHandle* HitHandle(Point3D p);
        
        /**
         * \brief Get the bounds of all selections in logical coordinates.
         */
        TDRectangle GetLocalBounds() const;
        
        virtual void StartDrag(Point3D currentPosition);
        
        virtual void Drag(Point3D currentPosition);
        
        virtual wxString EndDrag();
        
        /**
         * \brief Find the most specific common super class of all present
         *        selections.
         *
         * The returned class is selected such that all present selections
         * are instances of the class and for each subclass of the returned
         * class there is at least one selection in the current list which
         * is not an instance of the subclass.
         */
        wxClassInfo* FindCommonSelectionSuperClass();
        
        /**
         * \brief Find the first selection which is an instance of the given type.
         *
         * This is used to find a representative of all selections for providing
         * popup menu actions.
         */
        MapSelection* FindFirstSelectionOfType(wxClassInfo* classInfo);
        
        /**
         * \brief Find all selections which are an instance of the given type.
         */
        void GetSelectionsOfType(MapSelectionList& list, wxClassInfo* classInfo);
        
        /**
         * \brief Get the list of all selections.
         */
        const MapSelectionList& GetSelections() const { return _selections; }
        
        /**
         * \brief Get the MapEditor associated with this MapSelectionManager.
         */
        MapEditor* GetEditor() const { return _editor; }

        /**
         * \brief Register a MapSelectionChangeListener.
         *
         * A MapSelectionChangeListener is informed when the list of selections
         * changes.
         */
        void RegisterSelectionChangeListener(MapSelectionChangeListener* listener) {
                TransactionObject::Ref();
                _listMapSelectionChangeListener.Append(listener);
        }
        
        /**
         * \brief Unregister a MapSelectionChangeListener.
         */
        void UnregisterSelectionChangeListener(MapSelectionChangeListener* listener) {
                _listMapSelectionChangeListener.DeleteObject(listener);
                TransactionObject::UnRef();
        }
protected:
        void FireSelectionsAdded(const MapSelectionList& sellist) {
                MapSelectionChangeListenerList::Node *node;
                for (node=_listMapSelectionChangeListener.GetFirst();
                     node;
                     node=node->GetNext())
                        node->GetData()->SelectionsAdded(this,sellist);
        }
        
        void FireSelectionsRemoved(const MapSelectionList& sellist) {
                MapSelectionChangeListenerList::Node *node;
                for (node=_listMapSelectionChangeListener.GetFirst();
                     node;
                     node=node->GetNext())
                        node->GetData()->SelectionsRemoved(this,sellist);
        }
        
        class Memento: public wxObject {
                friend class MapSelectionManager;
        public:
                Memento(const MapSelectionManager& mgr) {
                        _selections=mgr._selections;
                        _selectionMap=mgr._selectionMap;
                        
                        MapSelectionList::Node* node;
                        for (node=_selections.GetFirst();node;node=node->GetNext())
                                node->GetData()->Ref();
                }
                
                virtual ~Memento() {
                        MapSelectionList::Node* node;
                        for (node=_selections.GetFirst();node;node=node->GetNext())
                                node->GetData()->UnRef();
                }
        protected:
                MapSelectionList _selections;
                MapSelectionMap _selectionMap;
        };
public:
        virtual Memento* CreateMemento() {
                return new Memento(*this);
        }
        
        virtual void RestoreMemento(wxObject* object);
protected:
        MapSelectionList _selections;
        MapSelectionMap _selectionMap;
        MapEditor* _editor;
        MapSelectionChangeListenerList _listMapSelectionChangeListener;
};

#endif

