#ifndef TransactionManager_h
#define TransactionManager_h

// For compilers that support precompilation, includes "wx.h".
#include <wx/wxprec.h>

#ifdef __BORLANDC__
#pragma hdrstop
#endif

#ifndef WX_PRECOMP
// Include your minimal set of headers here, or wx.h
#include <wx/wx.h>
#endif

#include "TransactionStackChangeListener.h"

#include <deque>

class TransactionObject;
class Transaction;

WX_DECLARE_LIST(TransactionStackChangeListener,TransactionStackChangeListenerList);

/**
 * \brief A TransactionManager provides services to undo and redo changes.
 *
 * The TransactionManager keeps a stack of Transaction instances.
 *
 * \invariant
 * The stack pointer always points to the index where a new transaction would be
 * placed.
 *
 * This means that if there is a transaction at the index of the stack pointer,
 * restoring that transaction will do nothing, as this transaction's mementos
 * represent the current state.
 *
 * Therefore on an undo action the transaction below the stack pointer must be
 * restored, while on an redo action the transaction above the stack pointer
 * must be restored.
 *
 * Transactions never store all changes, but only the changes occurring between
 * the end state of a transaction and the end state of its predecessor.
 *
 * Therefore on undo it is necessary to store copies of restored objects
 * in the successor transaction. When the stack pointer points to the top of stack
 * a backup transaction is created on top of the stack on undo.
 *
 * See \ref transaction_concept for more information.
 *
 * \todo We can simplify this class by ensuring that an uncommitted transaction
 *       is always on top of the stack.
 *
 * \sa Transaction TransactionObject
 */
class TransactionManager: public wxObject {
        
        DECLARE_CLASS(TransactionManager)
        
        /**
         * \todo Use portable wxWidgets types for this.
         */
        typedef std::deque<Transaction*> TransactionStack;
        typedef TransactionStack::size_type TransactionStackPtr, TransactionStackSize;
public:
	TransactionManager(TransactionStackSize backlogSize=100);
        virtual ~TransactionManager();
        
        /**
         * \brief Clear the transaction stack and reset the manager.
         *
         * This is used when a file is saved.
         */
        void Clear();
        
        /**
         * \brief Store a copy of the given object in the current transaction.
         *
         * This is typically called by a TransactionObject when it is about
         * to be modified for the first time in a transaction.
         */
        void StoreCopy(TransactionObject* obj);
        
        /**
         * \brief Delete the copies of the given object from all transactions.
         *
         * This is called by a TransactionObject when it is about to be destroyed.
         */
        void DeleteCopy(TransactionObject* obj);
        
        /**
         * \brief Commit the current transaction.
         *
         * \param description   Text to be used to describe the transaction
         *                      (e.g. in the undo and redo menu)
         */
        void Commit(const wxString& description=wxT(""));
        
        /**
         * \brief Abort the current transaction.
         *
         * All objects modified during the current transaction are restored
         * to their previous state.
         */
        void Abort();
        
        /**
         * \brief Undo one step.
         */
        void Undo();
        
        /**
         * \brief Redo one step.
         */
        void Redo();
        
        /**
         * \brief Determine whether at least one step can be undone.
         */
        bool CanUndo() const { return GetUndoSteps()>0; }
        
        /**
         * \brief Determine whether at least one step can be redone.
         */
        bool CanRedo() const { return GetRedoSteps()>0; }
        
        /**
         * \brief Determine the number of available undo steps.
         */
        TransactionStackSize GetUndoSteps() const {
                return _stackPtr;
        }
        
        /**
         * \brief Determine the number of available redo steps.
         */
        TransactionStackSize GetRedoSteps() const {
                /*
                 * While we are using the description from transaction
                 * _stackPtr to describe the redone changes, we actually
                 * restore transaction _stackPtr+1, which means that
                 * we only have _stack.size()-_stackPtr-1 steps,
                 * not _stack.size()-_stackPtr.
                 */
                if (_stackPtr==_stack.size())
                        return 0;
                return _stack.size()-_stackPtr-1;
        }
        
        /**
         * \brief Retrieve the description of the changes to be undone.
         *
         * \param idx   The index of the undo step, ranging from 0
         *              for the first to GetUndoSteps()-1 for the last step.
         */
        const wxString& GetUndoDescription(size_t idx=0) const;
        
        
        /**
         * \brief Retrieve the description of the changes to be redone.
         *
         * \param idx   The index of the undo step, ranging from 0
         *              for the first to GetRedoSteps()-1 for the last step.
         */
        const wxString& GetRedoDescription(size_t idx=0) const;
        
        /**
         * \brief Register a TransactionStackChangeListener.
         */
        void RegisterStackChangeListener(TransactionStackChangeListener* listener) {
                _transactionChangeListeners.Append(listener);
        }
        
        /**
         * \brief Unregister a TransactionStackChangeListener.
         */
        void UnregisterStackChangeListener(TransactionStackChangeListener* listener) {
                _transactionChangeListeners.DeleteObject(listener);
        }
        
        void Hold() { _holdCount++; }
        void Unhold() { if (--_holdCount==0) delete this; }
        
        unsigned GetHoldCount() const { return _holdCount; }
	
	void AddEditor() { _editorCount++; }
	void RemoveEditor() { _editorCount--; }
	
	unsigned int GetEditorCount() const { return _editorCount; }
protected:
        void OpenNewTransaction();
        
        void DestroyAllFrom(TransactionStackPtr start);
        
        void FireStackChange();
        
        TransactionStack _stack;
        
        /**
         * \brief The transaction stack pointer.
         */
        TransactionStackPtr _stackPtr;

        TransactionStackChangeListenerList _transactionChangeListeners;
        
        /**
         * \brief The maximum number of undoable transactions to keep
         *
         * If this is not zero, the TransactionManager will dispose of the
         * oldest transaction when a new transaction is opened and there
         * are already as many transactions as given by this value.
         *
         * \todo Make this settable 
         */
        TransactionStackSize _maxBacklogSize;
        
        unsigned int _holdCount,_editorCount;
};

#endif

